@@ -6,26 +6,14 @@
 #ifndef _EXT_OPENCL_H
 #define _EXT_OPENCL_H
 
+#define CL_TARGET_OPENCL_VERSION 120
+
 #define CL_USE_DEPRECATED_OPENCL_1_2_APIS
 #define CL_USE_DEPRECATED_OPENCL_2_0_APIS
 
 #if defined (__APPLE__)
 #include <OpenCL/cl.h>
-#endif
-
-#if defined (_WIN)
-#include <CL/cl.h>
-#endif
-
-#if defined (__linux__)
-#include <CL/cl.h>
-#endif
-
-#if defined (__FreeBSD__)
-#include <CL/cl.h>
-#endif
-
-#if defined (__CYGWIN__)
+#else
 #include <CL/cl.h>
 #endif
 
@@ -50,6 +38,7 @@ typedef union
 #define CL_PLATFORMS_MAX 16
 
 typedef cl_int           (CL_API_CALL *OCL_CLBUILDPROGRAM)            (cl_program, cl_uint, const cl_device_id *, const char *, void (CL_CALLBACK *)(cl_program, void *), void *);
+typedef cl_int           (CL_API_CALL *OCL_CLCOMPILEPROGRAM)          (cl_program, cl_uint, const cl_device_id *, const char *, cl_uint, const cl_program *, const char **, void (CL_CALLBACK *)(cl_program, void *), void *);
 typedef cl_mem           (CL_API_CALL *OCL_CLCREATEBUFFER)            (cl_context, cl_mem_flags, size_t, void *, cl_int *);
 typedef cl_command_queue (CL_API_CALL *OCL_CLCREATECOMMANDQUEUE)      (cl_context, cl_device_id, cl_command_queue_properties, cl_int *);
 typedef cl_context       (CL_API_CALL *OCL_CLCREATECONTEXT)           (const cl_context_properties *, cl_uint, const cl_device_id *, void (CL_CALLBACK *)(const char *, const void *, size_t, void *), void *, cl_int *);
@@ -73,6 +62,7 @@ typedef cl_int           (CL_API_CALL *OCL_CLGETPLATFORMIDS)          (cl_uint,
 typedef cl_int           (CL_API_CALL *OCL_CLGETPLATFORMINFO)         (cl_platform_id, cl_platform_info, size_t, void *, size_t *);
 typedef cl_int           (CL_API_CALL *OCL_CLGETPROGRAMBUILDINFO)     (cl_program, cl_device_id, cl_program_build_info, size_t, void *, size_t *);
 typedef cl_int           (CL_API_CALL *OCL_CLGETPROGRAMINFO)          (cl_program, cl_program_info, size_t, void *, size_t *);
+typedef cl_program       (CL_API_CALL *OCL_CLLINKPROGRAM)             (cl_context, cl_uint, const cl_device_id *, const char *, cl_uint, const cl_program *, void (CL_CALLBACK *) (cl_program, void *), void *, cl_int *);
 typedef cl_int           (CL_API_CALL *OCL_CLRELEASECOMMANDQUEUE)     (cl_command_queue);
 typedef cl_int           (CL_API_CALL *OCL_CLRELEASECONTEXT)          (cl_context);
 typedef cl_int           (CL_API_CALL *OCL_CLRELEASEEVENT)            (cl_event);
@@ -80,19 +70,15 @@ typedef cl_int           (CL_API_CALL *OCL_CLRELEASEKERNEL)           (cl_kernel
 typedef cl_int           (CL_API_CALL *OCL_CLRELEASEMEMOBJECT)        (cl_mem);
 typedef cl_int           (CL_API_CALL *OCL_CLRELEASEPROGRAM)          (cl_program);
 typedef cl_int           (CL_API_CALL *OCL_CLSETKERNELARG)            (cl_kernel, cl_uint, size_t, const void *);
+typedef cl_int           (CL_API_CALL *OCL_CLUNLOADPLATFORMCOMPILER)  (cl_platform_id);
 typedef cl_int           (CL_API_CALL *OCL_CLWAITFOREVENTS)           (cl_uint, const cl_event *);
 
-#if defined (_POSIX)
-typedef void *OCL_LIB;
-#else
-typedef HINSTANCE OCL_LIB;
-#endif
-
 typedef struct hc_opencl_lib
 {
-  OCL_LIB lib;
+  hc_dynlib_t lib;
 
   OCL_CLBUILDPROGRAM            clBuildProgram;
+  OCL_CLCOMPILEPROGRAM          clCompileProgram;
   OCL_CLCREATEBUFFER            clCreateBuffer;
   OCL_CLCREATECOMMANDQUEUE      clCreateCommandQueue;
   OCL_CLCREATECONTEXT           clCreateContext;
@@ -116,6 +102,7 @@ typedef struct hc_opencl_lib
   OCL_CLGETPLATFORMINFO         clGetPlatformInfo;
   OCL_CLGETPROGRAMBUILDINFO     clGetProgramBuildInfo;
   OCL_CLGETPROGRAMINFO          clGetProgramInfo;
+  OCL_CLLINKPROGRAM             clLinkProgram;
   OCL_CLRELEASECOMMANDQUEUE     clReleaseCommandQueue;
   OCL_CLRELEASECONTEXT          clReleaseContext;
   OCL_CLRELEASEEVENT            clReleaseEvent;
@@ -123,6 +110,7 @@ typedef struct hc_opencl_lib
   OCL_CLRELEASEMEMOBJECT        clReleaseMemObject;
   OCL_CLRELEASEPROGRAM          clReleaseProgram;
   OCL_CLSETKERNELARG            clSetKernelArg;
+  OCL_CLUNLOADPLATFORMCOMPILER  clUnloadPlatformCompiler;
   OCL_CLWAITFOREVENTS           clWaitForEvents;
 
 } hc_opencl_lib_t;
