@@ -8,20 +8,18 @@
 #include "memory.h"
 #include "bitops.h"
 #include "event.h"
-#include "dictstat.h"
 #include "locking.h"
 #include "shared.h"
+#include "dictstat.h"
 
 int sort_by_dictstat (const void *s1, const void *s2)
 {
-  dictstat_t *d1 = (dictstat_t *) s1;
-  dictstat_t *d2 = (dictstat_t *) s2;
+  const dictstat_t *d1 = (const dictstat_t *) s1;
+  const dictstat_t *d2 = (const dictstat_t *) s2;
 
-  d2->stat.st_atime = d1->stat.st_atime;
+  const int rc_hash = memcmp (d1->hash_filename, d2->hash_filename, 16);
 
-  #if defined (STAT_NANOSECONDS_ACCESS_TIME)
-  d2->stat.STAT_NANOSECONDS_ACCESS_TIME = d1->stat.STAT_NANOSECONDS_ACCESS_TIME;
-  #endif
+  if (rc_hash != 0) return rc_hash;
 
   const int rc_from = strcmp (d1->encoding_from, d2->encoding_from);
 
@@ -31,7 +29,23 @@ int sort_by_dictstat (const void *s1, const void *s2)
 
   if (rc_to != 0) return rc_to;
 
-  return memcmp (&d1->stat, &d2->stat, sizeof (struct stat));
+  struct stat stat1;
+  struct stat stat2;
+
+  memcpy (&stat1, &d1->stat, sizeof (struct stat));
+  memcpy (&stat2, &d2->stat, sizeof (struct stat));
+
+  stat1.st_atime = 0;
+  stat2.st_atime = 0;
+
+  #if defined (STAT_NANOSECONDS_ACCESS_TIME)
+  stat1.STAT_NANOSECONDS_ACCESS_TIME = 0;
+  stat2.STAT_NANOSECONDS_ACCESS_TIME = 0;
+  #endif
+
+  const int rc_memcmp = memcmp (&stat1, &stat2, sizeof (struct stat));
+
+  return rc_memcmp;
 }
 
 int dictstat_init (hashcat_ctx_t *hashcat_ctx)
@@ -43,18 +57,16 @@ int dictstat_init (hashcat_ctx_t *hashcat_ctx)
   dictstat_ctx->enabled = false;
 
   if (user_options->benchmark      == true) return 0;
-  if (user_options->example_hashes == true) return 0;
+  if (user_options->hash_info      == true) return 0;
   if (user_options->keyspace       == true) return 0;
   if (user_options->left           == true) return 0;
-  if (user_options->opencl_info    == true) return 0;
+  if (user_options->backend_info   == true) return 0;
   if (user_options->show           == true) return 0;
   if (user_options->usage          == true) return 0;
   if (user_options->version        == true) return 0;
 
   if (user_options->attack_mode == ATTACK_MODE_BF) return 0;
 
-  if (user_options->hash_mode == 3000) return 0; // this mode virtually creates words in the wordlists
-
   dictstat_ctx->enabled  = true;
   dictstat_ctx->base     = (dictstat_t *) hccalloc (MAX_DICTSTAT, sizeof (dictstat_t));
   dictstat_ctx->cnt      = 0;
@@ -66,10 +78,13 @@ int dictstat_init (hashcat_ctx_t *hashcat_ctx)
 
 void dictstat_destroy (hashcat_ctx_t *hashcat_ctx)
 {
+  hashconfig_t   *hashconfig   = hashcat_ctx->hashconfig;
   dictstat_ctx_t *dictstat_ctx = hashcat_ctx->dictstat_ctx;
 
   if (dictstat_ctx->enabled == false) return;
 
+  if (hashconfig->dictstat_disable == true) return;
+
   hcfree (dictstat_ctx->filename);
   hcfree (dictstat_ctx->base);
 
@@ -78,13 +93,16 @@ void dictstat_destroy (hashcat_ctx_t *hashcat_ctx)
 
 void dictstat_read (hashcat_ctx_t *hashcat_ctx)
 {
+  hashconfig_t   *hashconfig   = hashcat_ctx->hashconfig;
   dictstat_ctx_t *dictstat_ctx = hashcat_ctx->dictstat_ctx;
 
   if (dictstat_ctx->enabled == false) return;
 
-  FILE *fp = fopen (dictstat_ctx->filename, "rb");
+  if (hashconfig->dictstat_disable == true) return;
 
-  if (fp == NULL)
+  HCFILE fp;
+
+  if (hc_fopen (&fp, dictstat_ctx->filename, "rb") == false)
   {
     // first run, file does not exist, do not error out
 
@@ -96,14 +114,14 @@ void dictstat_read (hashcat_ctx_t *hashcat_ctx)
   u64 v;
   u64 z;
 
-  const size_t nread1 = hc_fread (&v, sizeof (u64), 1, fp);
-  const size_t nread2 = hc_fread (&z, sizeof (u64), 1, fp);
+  const size_t nread1 = hc_fread (&v, sizeof (u64), 1, &fp);
+  const size_t nread2 = hc_fread (&z, sizeof (u64), 1, &fp);
 
   if ((nread1 != 1) || (nread2 != 1))
   {
     event_log_error (hashcat_ctx, "%s: Invalid header", dictstat_ctx->filename);
 
-    fclose (fp);
+    hc_fclose (&fp);
 
     return;
   }
@@ -115,7 +133,7 @@ void dictstat_read (hashcat_ctx_t *hashcat_ctx)
   {
     event_log_error (hashcat_ctx, "%s: Invalid header, ignoring content", dictstat_ctx->filename);
 
-    fclose (fp);
+    hc_fclose (&fp);
 
     return;
   }
@@ -124,7 +142,7 @@ void dictstat_read (hashcat_ctx_t *hashcat_ctx)
   {
     event_log_error (hashcat_ctx, "%s: Invalid header, ignoring content", dictstat_ctx->filename);
 
-    fclose (fp);
+    hc_fclose (&fp);
 
     return;
   }
@@ -133,18 +151,18 @@ void dictstat_read (hashcat_ctx_t *hashcat_ctx)
   {
     event_log_warning (hashcat_ctx, "%s: Outdated header version, ignoring content", dictstat_ctx->filename);
 
-    fclose (fp);
+    hc_fclose (&fp);
 
     return;
   }
 
   // parse data
 
-  while (!feof (fp))
+  while (!hc_feof (&fp))
   {
     dictstat_t d;
 
-    const size_t nread = hc_fread (&d, sizeof (dictstat_t), 1, fp);
+    const size_t nread = hc_fread (&d, sizeof (dictstat_t), 1, &fp);
 
     if (nread == 0) continue;
 
@@ -158,27 +176,30 @@ void dictstat_read (hashcat_ctx_t *hashcat_ctx)
     }
   }
 
-  fclose (fp);
+  hc_fclose (&fp);
 }
 
 int dictstat_write (hashcat_ctx_t *hashcat_ctx)
 {
+  hashconfig_t   *hashconfig   = hashcat_ctx->hashconfig;
   dictstat_ctx_t *dictstat_ctx = hashcat_ctx->dictstat_ctx;
 
   if (dictstat_ctx->enabled == false) return 0;
 
-  FILE *fp = fopen (dictstat_ctx->filename, "wb");
+  if (hashconfig->dictstat_disable == true) return 0;
 
-  if (fp == NULL)
+  HCFILE fp;
+
+  if (hc_fopen (&fp, dictstat_ctx->filename, "wb") == false)
   {
     event_log_error (hashcat_ctx, "%s: %s", dictstat_ctx->filename, strerror (errno));
 
     return -1;
   }
 
-  if (lock_file (fp) == -1)
+  if (hc_lockfile (&fp) == -1)
   {
-    fclose (fp);
+    hc_fclose (&fp);
 
     event_log_error (hashcat_ctx, "%s: %s", dictstat_ctx->filename, strerror (errno));
 
@@ -193,24 +214,36 @@ int dictstat_write (hashcat_ctx_t *hashcat_ctx)
   v = byte_swap_64 (v);
   z = byte_swap_64 (z);
 
-  hc_fwrite (&v, sizeof (u64), 1, fp);
-  hc_fwrite (&z, sizeof (u64), 1, fp);
+  hc_fwrite (&v, sizeof (u64), 1, &fp);
+  hc_fwrite (&z, sizeof (u64), 1, &fp);
 
   // data
 
-  hc_fwrite (dictstat_ctx->base, sizeof (dictstat_t), dictstat_ctx->cnt, fp);
+  hc_fwrite (dictstat_ctx->base, sizeof (dictstat_t), dictstat_ctx->cnt, &fp);
 
-  fclose (fp);
+  if (hc_unlockfile (&fp) == -1)
+  {
+    hc_fclose (&fp);
+
+    event_log_error (hashcat_ctx, "%s: %s", dictstat_ctx->filename, strerror (errno));
+
+    return -1;
+  }
+
+  hc_fclose (&fp);
 
   return 0;
 }
 
 u64 dictstat_find (hashcat_ctx_t *hashcat_ctx, dictstat_t *d)
 {
+  hashconfig_t   *hashconfig   = hashcat_ctx->hashconfig;
   dictstat_ctx_t *dictstat_ctx = hashcat_ctx->dictstat_ctx;
 
   if (dictstat_ctx->enabled == false) return 0;
 
+  if (hashconfig->dictstat_disable == true) return 0;
+
   dictstat_t *d_cache = (dictstat_t *) lfind (d, dictstat_ctx->base, &dictstat_ctx->cnt, sizeof (dictstat_t), sort_by_dictstat);
 
   if (d_cache == NULL) return 0;
@@ -220,10 +253,13 @@ u64 dictstat_find (hashcat_ctx_t *hashcat_ctx, dictstat_t *d)
 
 void dictstat_append (hashcat_ctx_t *hashcat_ctx, dictstat_t *d)
 {
+  hashconfig_t   *hashconfig   = hashcat_ctx->hashconfig;
   dictstat_ctx_t *dictstat_ctx = hashcat_ctx->dictstat_ctx;
 
   if (dictstat_ctx->enabled == false) return;
 
+  if (hashconfig->dictstat_disable == true) return;
+
   if (dictstat_ctx->cnt == MAX_DICTSTAT)
   {
     event_log_error (hashcat_ctx, "There are too many entries in the %s database. You have to remove/rename it.", dictstat_ctx->filename);
