@@ -5,10 +5,9 @@
 
 #include "common.h"
 #include "types.h"
-#include "convert.h"
 #include "event.h"
 #include "memory.h"
-#include "opencl.h"
+#include "backend.h"
 #include "wordlist.h"
 #include "shared.h"
 #include "thread.h"
@@ -16,7 +15,6 @@
 #include "rp.h"
 #include "rp_cpu.h"
 #include "slow_candidates.h"
-#include "user_options.h"
 #include "dispatch.h"
 
 #ifdef WITH_BRAIN
@@ -25,16 +23,18 @@
 
 static u64 get_highest_words_done (const hashcat_ctx_t *hashcat_ctx)
 {
-  const opencl_ctx_t *opencl_ctx = hashcat_ctx->opencl_ctx;
+  const backend_ctx_t *backend_ctx = hashcat_ctx->backend_ctx;
 
   u64 words_cur = 0;
 
-  for (u32 device_id = 0; device_id < opencl_ctx->devices_cnt; device_id++)
+  for (int backend_devices_idx = 0; backend_devices_idx < backend_ctx->backend_devices_cnt; backend_devices_idx++)
   {
-    hc_device_param_t *device_param = &opencl_ctx->devices_param[device_id];
+    hc_device_param_t *device_param = &backend_ctx->devices_param[backend_devices_idx];
 
     if (device_param->skipped == true) continue;
 
+    if (device_param->skipped_warning == true) continue;
+
     const u64 words_done = device_param->words_done;
 
     if (words_done > words_cur) words_cur = words_done;
@@ -45,16 +45,18 @@ static u64 get_highest_words_done (const hashcat_ctx_t *hashcat_ctx)
 
 static u64 get_lowest_words_done (const hashcat_ctx_t *hashcat_ctx)
 {
-  const opencl_ctx_t *opencl_ctx = hashcat_ctx->opencl_ctx;
+  const backend_ctx_t *backend_ctx = hashcat_ctx->backend_ctx;
 
   u64 words_cur = 0xffffffffffffffff;
 
-  for (u32 device_id = 0; device_id < opencl_ctx->devices_cnt; device_id++)
+  for (int backend_devices_idx = 0; backend_devices_idx < backend_ctx->backend_devices_cnt; backend_devices_idx++)
   {
-    hc_device_param_t *device_param = &opencl_ctx->devices_param[device_id];
+    hc_device_param_t *device_param = &backend_ctx->devices_param[backend_devices_idx];
 
     if (device_param->skipped == true) continue;
 
+    if (device_param->skipped_warning == true) continue;
+
     const u64 words_done = device_param->words_done;
 
     if (words_done < words_cur) words_cur = words_done;
@@ -74,20 +76,20 @@ static int set_kernel_power_final (hashcat_ctx_t *hashcat_ctx, const u64 kernel_
 {
   EVENT (EVENT_SET_KERNEL_POWER_FINAL);
 
-  opencl_ctx_t *opencl_ctx = hashcat_ctx->opencl_ctx;
+  backend_ctx_t *backend_ctx = hashcat_ctx->backend_ctx;
 
-  opencl_ctx->kernel_power_final = kernel_power_final;
+  backend_ctx->kernel_power_final = kernel_power_final;
 
   return 0;
 }
 
-static u64 get_power (opencl_ctx_t *opencl_ctx, hc_device_param_t *device_param)
+static u64 get_power (backend_ctx_t *backend_ctx, hc_device_param_t *device_param)
 {
-  const u64 kernel_power_final = opencl_ctx->kernel_power_final;
+  const u64 kernel_power_final = backend_ctx->kernel_power_final;
 
   if (kernel_power_final)
   {
-    const double device_factor = (double) device_param->hardware_power / opencl_ctx->hardware_power_all;
+    const double device_factor = (double) device_param->hardware_power / backend_ctx->hardware_power_all;
 
     const u64 words_left_device = (u64) CEIL (kernel_power_final * device_factor);
 
@@ -107,7 +109,7 @@ static u64 get_power (opencl_ctx_t *opencl_ctx, hc_device_param_t *device_param)
 
 static u64 get_work (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param, const u64 max)
 {
-  opencl_ctx_t   *opencl_ctx   = hashcat_ctx->opencl_ctx;
+  backend_ctx_t  *backend_ctx  = hashcat_ctx->backend_ctx;
   status_ctx_t   *status_ctx   = hashcat_ctx->status_ctx;
   user_options_t *user_options = hashcat_ctx->user_options;
 
@@ -118,19 +120,19 @@ static u64 get_work (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param
 
   device_param->words_off = words_off;
 
-  const u64 kernel_power_all = opencl_ctx->kernel_power_all;
+  const u64 kernel_power_all = backend_ctx->kernel_power_all;
 
   const u64 words_left = words_base - words_off;
 
   if (words_left < kernel_power_all)
   {
-    if (opencl_ctx->kernel_power_final == 0)
+    if (backend_ctx->kernel_power_final == 0)
     {
       set_kernel_power_final (hashcat_ctx, words_left);
     }
   }
 
-  const u64 kernel_power = get_power (opencl_ctx, device_param);
+  const u64 kernel_power = get_power (backend_ctx, device_param);
 
   u64 work = MIN (words_left, kernel_power);
 
@@ -166,7 +168,12 @@ static int calc_stdin (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_par
 
     iconv_ctx = iconv_open (user_options->encoding_to, user_options->encoding_from);
 
-    if (iconv_ctx == (iconv_t) -1) return -1;
+    if (iconv_ctx == (iconv_t) -1)
+    {
+      hcfree (buf);
+
+      return -1;
+    }
 
     iconv_tmp = (char *) hcmalloc (HCBUFSIZ_TINY);
   }
@@ -219,9 +226,7 @@ static int calc_stdin (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_par
         char  *iconv_ptr = iconv_tmp;
         size_t iconv_sz  = HCBUFSIZ_TINY;
 
-        const size_t iconv_rc = iconv (iconv_ctx, &line_buf, &line_len, &iconv_ptr, &iconv_sz);
-
-        if (iconv_rc == (size_t) -1) continue;
+        if (iconv (iconv_ctx, &line_buf, &line_len, &iconv_ptr, &iconv_sz) == (size_t) -1) continue;
 
         line_buf = iconv_tmp;
         line_len = HCBUFSIZ_TINY - iconv_sz;
@@ -245,7 +250,7 @@ static int calc_stdin (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_par
         line_len = (size_t) rule_len_out;
       }
 
-      if (line_len >= PW_MAX) continue;
+      if (line_len > PW_MAX) continue;
 
       // hmm that's always the case, or?
 
@@ -286,20 +291,14 @@ static int calc_stdin (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_par
 
     // flush
 
-    int CL_rc;
-
-    CL_rc = run_copy (hashcat_ctx, device_param, device_param->pws_cnt);
-
-    if (CL_rc == -1)
+    if (run_copy (hashcat_ctx, device_param, device_param->pws_cnt) == -1)
     {
       hcfree (buf);
 
       return -1;
     }
 
-    CL_rc = run_cracker (hashcat_ctx, device_param, device_param->pws_cnt);
-
-    if (CL_rc == -1)
+    if (run_cracker (hashcat_ctx, device_param, -1, device_param->pws_cnt) == -1) // no pws_pos?
     {
       hcfree (buf);
 
@@ -336,18 +335,22 @@ HC_API_CALL void *thread_calc_stdin (void *p)
   thread_param_t *thread_param = (thread_param_t *) p;
 
   hashcat_ctx_t *hashcat_ctx = thread_param->hashcat_ctx;
+  backend_ctx_t *backend_ctx = hashcat_ctx->backend_ctx;
 
-  opencl_ctx_t *opencl_ctx = hashcat_ctx->opencl_ctx;
+  if (backend_ctx->enabled == false) return NULL;
 
-  if (opencl_ctx->enabled == false) return NULL;
-
-  hc_device_param_t *device_param = opencl_ctx->devices_param + thread_param->tid;
+  hc_device_param_t *device_param = backend_ctx->devices_param + thread_param->tid;
 
   if (device_param->skipped) return NULL;
 
-  const int rc_calc = calc_stdin (hashcat_ctx, device_param);
+  if (device_param->skipped_warning == true) return NULL;
 
-  if (rc_calc == -1)
+  if (device_param->is_cuda == true)
+  {
+    if (hc_cuCtxSetCurrent (hashcat_ctx, device_param->cuda_context) == -1) return NULL;
+  }
+
+  if (calc_stdin (hashcat_ctx, device_param) == -1)
   {
     status_ctx_t *status_ctx = hashcat_ctx->status_ctx;
 
@@ -366,7 +369,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
   mask_ctx_t           *mask_ctx           = hashcat_ctx->mask_ctx;
   straight_ctx_t       *straight_ctx       = hashcat_ctx->straight_ctx;
   combinator_ctx_t     *combinator_ctx     = hashcat_ctx->combinator_ctx;
-  opencl_ctx_t         *opencl_ctx         = hashcat_ctx->opencl_ctx;
+  backend_ctx_t        *backend_ctx        = hashcat_ctx->backend_ctx;
   status_ctx_t         *status_ctx         = hashcat_ctx->status_ctx;
 
   const u32 attack_mode = user_options->attack_mode;
@@ -391,19 +394,22 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
         brain_client_disconnect (device_param);
       }
 
-      hc_thread_mutex_lock (status_ctx->mux_dispatcher);
-
-      if (status_ctx->words_off == 0)
+      if (user_options->brain_client_features & BRAIN_CLIENT_FEATURE_ATTACKS)
       {
-        status_ctx->words_off = highest;
+        hc_thread_mutex_lock (status_ctx->mux_dispatcher);
 
-        for (u32 salt_pos = 0; salt_pos < hashes->salts_cnt; salt_pos++)
+        if (status_ctx->words_off == 0)
         {
-          status_ctx->words_progress_rejected[salt_pos] = status_ctx->words_off;
-        }
-      }
+          status_ctx->words_off = highest;
 
-      hc_thread_mutex_unlock (status_ctx->mux_dispatcher);
+          for (u32 salt_pos = 0; salt_pos < hashes->salts_cnt; salt_pos++)
+          {
+            status_ctx->words_progress_rejected[salt_pos] = status_ctx->words_off;
+          }
+        }
+
+        hc_thread_mutex_unlock (status_ctx->mux_dispatcher);
+      }
     }
     #endif
 
@@ -413,32 +419,26 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
     {
       char *dictfile = straight_ctx->dict;
 
-      FILE *fp = fopen (dictfile, "rb");
+      extra_info_straight_t extra_info_straight;
 
-      if (fp == NULL)
+      memset (&extra_info_straight, 0, sizeof (extra_info_straight));
+
+      if (hc_fopen (&extra_info_straight.fp, dictfile, "rb") == false)
       {
         event_log_error (hashcat_ctx, "%s: %s", dictfile, strerror (errno));
 
         return -1;
       }
 
-      extra_info_straight_t extra_info_straight;
-
-      memset (&extra_info_straight, 0, sizeof (extra_info_straight));
-
-      extra_info_straight.fp = fp;
-
       hashcat_ctx_t *hashcat_ctx_tmp = (hashcat_ctx_t *) hcmalloc (sizeof (hashcat_ctx_t));
 
       memcpy (hashcat_ctx_tmp, hashcat_ctx, sizeof (hashcat_ctx_t)); // yes we actually want to copy these pointers
 
       hashcat_ctx_tmp->wl_data = (wl_data_t *) hcmalloc (sizeof (wl_data_t));
 
-      const int rc_wl_data_init = wl_data_init (hashcat_ctx_tmp);
-
-      if (rc_wl_data_init == -1)
+      if (wl_data_init (hashcat_ctx_tmp) == -1)
       {
-        fclose (fp);
+        hc_fclose (&extra_info_straight.fp);
 
         hcfree (hashcat_ctx_tmp->wl_data);
 
@@ -461,7 +461,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
         // this greatly reduces spam on hashcat console
 
-        const u64 pre_rejects_ignore = get_power (opencl_ctx, device_param) / 2;
+        const u64 pre_rejects_ignore = get_power (backend_ctx, device_param) / 2;
 
         while (pre_rejects > pre_rejects_ignore)
         {
@@ -505,13 +505,10 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
                   brain_client_disconnect (device_param);
                 }
 
-                words_extra = overlap;
-
+                words_extra        = overlap;
                 words_extra_total += overlap;
-
-                words_off += overlap;
-
-                work -= overlap;
+                words_off         += overlap;
+                work              -= overlap;
               }
             }
             #endif
@@ -544,7 +541,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
                 brain_client_generate_hash ((u64 *) hash, (const char *) extra_info_straight.out_buf, extra_info_straight.out_len);
 
-                u32 *ptr = (u32 *) device_param->brain_link_out_buf;
+                u32 *ptr = device_param->brain_link_out_buf;
 
                 ptr[(device_param->pws_pre_cnt * 2) + 0] = hash[0];
                 ptr[(device_param->pws_pre_cnt * 2) + 1] = hash[1];
@@ -643,29 +640,21 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
         if (pws_cnt)
         {
-          int CL_rc;
-
-          CL_rc = run_copy (hashcat_ctx, device_param, pws_cnt);
-
-          if (CL_rc == -1)
+          if (run_copy (hashcat_ctx, device_param, pws_cnt) == -1)
           {
-            fclose (fp);
+            hc_fclose (&extra_info_straight.fp);
 
             hcfree (hashcat_ctx_tmp->wl_data);
-
             hcfree (hashcat_ctx_tmp);
 
             return -1;
           }
 
-          CL_rc = run_cracker (hashcat_ctx, device_param, pws_cnt);
-
-          if (CL_rc == -1)
+          if (run_cracker (hashcat_ctx, device_param, -1, pws_cnt) == -1)
           {
-            fclose (fp);
+            hc_fclose (&extra_info_straight.fp);
 
             hcfree (hashcat_ctx_tmp->wl_data);
-
             hcfree (hashcat_ctx_tmp);
 
             return -1;
@@ -688,8 +677,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
           }
           #endif
 
-          device_param->pws_cnt = 0;
-
+          device_param->pws_cnt      = 0;
           device_param->pws_base_cnt = 0;
         }
 
@@ -707,12 +695,11 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
         if (words_fin == 0) break;
       }
 
-      fclose (fp);
+      hc_fclose (&extra_info_straight.fp);
 
       wl_data_destroy (hashcat_ctx_tmp);
 
       hcfree (hashcat_ctx_tmp->wl_data);
-
       hcfree (hashcat_ctx_tmp);
     }
     else if (attack_mode == ATTACK_MODE_COMBI)
@@ -733,30 +720,26 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
         combs_file = combinator_ctx->dict1;
       }
 
-      FILE *base_fp = fopen (base_file, "rb");
+      extra_info_combi_t extra_info_combi;
 
-      if (base_fp == NULL)
+      memset (&extra_info_combi, 0, sizeof (extra_info_combi));
+
+      if (hc_fopen (&extra_info_combi.base_fp, base_file, "rb") == false)
       {
         event_log_error (hashcat_ctx, "%s: %s", base_file, strerror (errno));
 
         return -1;
       }
 
-      FILE *combs_fp = fopen (combs_file, "rb");
-
-      if (combs_fp == NULL)
+      if (hc_fopen (&extra_info_combi.combs_fp, combs_file, "rb") == false)
       {
         event_log_error (hashcat_ctx, "%s: %s", combs_file, strerror (errno));
 
+        hc_fclose (&extra_info_combi.base_fp);
+
         return -1;
       }
 
-      extra_info_combi_t extra_info_combi;
-
-      memset (&extra_info_combi, 0, sizeof (extra_info_combi));
-
-      extra_info_combi.base_fp     = base_fp;
-      extra_info_combi.combs_fp    = combs_fp;
       extra_info_combi.scratch_buf = device_param->scratch_buf;
 
       hashcat_ctx_t *hashcat_ctx_tmp = (hashcat_ctx_t *) hcmalloc (sizeof (hashcat_ctx_t));
@@ -765,16 +748,12 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
       hashcat_ctx_tmp->wl_data = (wl_data_t *) hcmalloc (sizeof (wl_data_t));
 
-      const int rc_wl_data_init = wl_data_init (hashcat_ctx_tmp);
-
-      if (rc_wl_data_init == -1)
+      if (wl_data_init (hashcat_ctx_tmp) == -1)
       {
-        fclose (combs_fp);
-
-        fclose (base_fp);
+        hc_fclose (&extra_info_combi.base_fp);
+        hc_fclose (&extra_info_combi.combs_fp);
 
         hcfree (hashcat_ctx_tmp->wl_data);
-
         hcfree (hashcat_ctx_tmp);
 
         return -1;
@@ -794,7 +773,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
         // this greatly reduces spam on hashcat console
 
-        const u64 pre_rejects_ignore = get_power (opencl_ctx, device_param) / 2;
+        const u64 pre_rejects_ignore = get_power (backend_ctx, device_param) / 2;
 
         while (pre_rejects > pre_rejects_ignore)
         {
@@ -840,13 +819,10 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
                   brain_client_disconnect (device_param);
                 }
 
-                words_extra = overlap;
-
+                words_extra        = overlap;
                 words_extra_total += overlap;
-
-                words_off += overlap;
-
-                work -= overlap;
+                words_off         += overlap;
+                work              -= overlap;
               }
             }
             #endif
@@ -877,7 +853,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
                 brain_client_generate_hash ((u64 *) hash, (const char *) extra_info_combi.out_buf, extra_info_combi.out_len);
 
-                u32 *ptr = (u32 *) device_param->brain_link_out_buf;
+                u32 *ptr = device_param->brain_link_out_buf;
 
                 ptr[(device_param->pws_pre_cnt * 2) + 0] = hash[0];
                 ptr[(device_param->pws_pre_cnt * 2) + 1] = hash[1];
@@ -976,33 +952,23 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
         if (pws_cnt)
         {
-          int CL_rc;
-
-          CL_rc = run_copy (hashcat_ctx, device_param, pws_cnt);
-
-          if (CL_rc == -1)
+          if (run_copy (hashcat_ctx, device_param, pws_cnt) == -1)
           {
-            fclose (combs_fp);
-
-            fclose (base_fp);
+            hc_fclose (&extra_info_combi.base_fp);
+            hc_fclose (&extra_info_combi.combs_fp);
 
             hcfree (hashcat_ctx_tmp->wl_data);
-
             hcfree (hashcat_ctx_tmp);
 
             return -1;
           }
 
-          CL_rc = run_cracker (hashcat_ctx, device_param, pws_cnt);
-
-          if (CL_rc == -1)
+          if (run_cracker (hashcat_ctx, device_param, -1, pws_cnt) == -1)
           {
-            fclose (combs_fp);
-
-            fclose (base_fp);
+            hc_fclose (&extra_info_combi.base_fp);
+            hc_fclose (&extra_info_combi.combs_fp);
 
             hcfree (hashcat_ctx_tmp->wl_data);
-
             hcfree (hashcat_ctx_tmp);
 
             return -1;
@@ -1025,8 +991,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
           }
           #endif
 
-          device_param->pws_cnt = 0;
-
+          device_param->pws_cnt      = 0;
           device_param->pws_base_cnt = 0;
         }
 
@@ -1044,14 +1009,12 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
         if (words_fin == 0) break;
       }
 
-      fclose (combs_fp);
-
-      fclose (base_fp);
+      hc_fclose (&extra_info_combi.base_fp);
+      hc_fclose (&extra_info_combi.combs_fp);
 
       wl_data_destroy (hashcat_ctx_tmp);
 
       hcfree (hashcat_ctx_tmp->wl_data);
-
       hcfree (hashcat_ctx_tmp);
     }
     else if (attack_mode == ATTACK_MODE_BF)
@@ -1075,7 +1038,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
         // this greatly reduces spam on hashcat console
 
-        const u64 pre_rejects_ignore = get_power (opencl_ctx, device_param) / 2;
+        const u64 pre_rejects_ignore = get_power (backend_ctx, device_param) / 2;
 
         while (pre_rejects > pre_rejects_ignore)
         {
@@ -1121,19 +1084,15 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
                   brain_client_disconnect (device_param);
                 }
 
-                words_extra = overlap;
-
+                words_extra        = overlap;
                 words_extra_total += overlap;
-
-                words_off += overlap;
-
-                work -= overlap;
+                words_off         += overlap;
+                work              -= overlap;
               }
             }
             #endif
 
             words_fin = words_off + work;
-
             words_cur = words_off;
 
             for (u64 i = words_cur; i < words_fin; i++)
@@ -1149,7 +1108,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
                 brain_client_generate_hash ((u64 *) hash, (const char *) extra_info_mask.out_buf, extra_info_mask.out_len);
 
-                u32 *ptr = (u32 *) device_param->brain_link_out_buf;
+                u32 *ptr = device_param->brain_link_out_buf;
 
                 ptr[(device_param->pws_pre_cnt * 2) + 0] = hash[0];
                 ptr[(device_param->pws_pre_cnt * 2) + 1] = hash[1];
@@ -1242,21 +1201,8 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
         if (pws_cnt)
         {
-          int CL_rc;
-
-          CL_rc = run_copy (hashcat_ctx, device_param, pws_cnt);
-
-          if (CL_rc == -1)
-          {
-            return -1;
-          }
-
-          CL_rc = run_cracker (hashcat_ctx, device_param, pws_cnt);
-
-          if (CL_rc == -1)
-          {
-            return -1;
-          }
+          if (run_copy    (hashcat_ctx, device_param, pws_cnt) == -1) return -1;
+          if (run_cracker (hashcat_ctx, device_param, -1, pws_cnt) == -1) return -1;
 
           #ifdef WITH_BRAIN
           if (user_options->brain_client == true)
@@ -1308,16 +1254,12 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
       {
         char *dictfile = straight_ctx->dict;
 
-        FILE *combs_fp = fopen (dictfile, "rb");
-
-        if (combs_fp == NULL)
+        if (hc_fopen (&device_param->combs_fp, dictfile, "rb") == false)
         {
           event_log_error (hashcat_ctx, "%s: %s", dictfile, strerror (errno));
 
           return -1;
         }
-
-        device_param->combs_fp = combs_fp;
       }
 
       while (status_ctx->run_thread_level1 == true)
@@ -1331,15 +1273,8 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
         device_param->pws_cnt = work;
 
-        int CL_rc;
-
-        CL_rc = run_copy (hashcat_ctx, device_param, device_param->pws_cnt);
-
-        if (CL_rc == -1) return -1;
-
-        CL_rc = run_cracker (hashcat_ctx, device_param, device_param->pws_cnt);
-
-        if (CL_rc == -1) return -1;
+        if (run_copy    (hashcat_ctx, device_param, device_param->pws_cnt) == -1) return -1;
+        if (run_cracker (hashcat_ctx, device_param, -1, device_param->pws_cnt) == -1) return -1;
 
         device_param->pws_cnt = 0;
 
@@ -1374,37 +1309,29 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
         {
           const char *dictfilec = combinator_ctx->dict2;
 
-          FILE *combs_fp = fopen (dictfilec, "rb");
-
-          if (combs_fp == NULL)
+          if (hc_fopen (&device_param->combs_fp, dictfilec, "rb") == false)
           {
             event_log_error (hashcat_ctx, "%s: %s", combinator_ctx->dict2, strerror (errno));
 
             return -1;
           }
-
-          device_param->combs_fp = combs_fp;
         }
         else if (combs_mode == COMBINATOR_MODE_BASE_RIGHT)
         {
           const char *dictfilec = combinator_ctx->dict1;
 
-          FILE *combs_fp = fopen (dictfilec, "rb");
-
-          if (combs_fp == NULL)
+          if (hc_fopen (&device_param->combs_fp, dictfilec, "rb") == false)
           {
             event_log_error (hashcat_ctx, "%s: %s", dictfilec, strerror (errno));
 
             return -1;
           }
-
-          device_param->combs_fp = combs_fp;
         }
       }
 
-      FILE *fd = fopen (dictfile, "rb");
+      HCFILE fp;
 
-      if (fd == NULL)
+      if (hc_fopen (&fp, dictfile, "rb") == false)
       {
         event_log_error (hashcat_ctx, "%s: %s", dictfile, strerror (errno));
 
@@ -1417,16 +1344,13 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
       hashcat_ctx_tmp->wl_data = (wl_data_t *) hcmalloc (sizeof (wl_data_t));
 
-      const int rc_wl_data_init = wl_data_init (hashcat_ctx_tmp);
-
-      if (rc_wl_data_init == -1)
+      if (wl_data_init (hashcat_ctx_tmp) == -1)
       {
-        if (attack_mode == ATTACK_MODE_COMBI) fclose (device_param->combs_fp);
+        if (attack_mode == ATTACK_MODE_COMBI) hc_fclose (&device_param->combs_fp);
 
-        fclose (fd);
+        hc_fclose (&fp);
 
         hcfree (hashcat_ctx_tmp->wl_data);
-
         hcfree (hashcat_ctx_tmp);
 
         return -1;
@@ -1438,9 +1362,7 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
       {
         u64 words_off = 0;
         u64 words_fin = 0;
-
-        u64 words_extra = -1u;
-
+        u64 words_extra = -1U;
         u64 words_extra_total = 0;
 
         memset (device_param->pws_comp, 0, device_param->size_pws_comp);
@@ -1462,13 +1384,11 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
           char rule_buf_out[RP_PASSWORD_SIZE];
 
-          for ( ; words_cur < words_off; words_cur++) get_next_word (hashcat_ctx_tmp, fd, &line_buf, &line_len);
+          for ( ; words_cur < words_off; words_cur++) get_next_word (hashcat_ctx_tmp, &fp, &line_buf, &line_len);
 
           for ( ; words_cur < words_fin; words_cur++)
           {
-            get_next_word (hashcat_ctx_tmp, fd, &line_buf, &line_len);
-
-            line_len = (u32) convert_from_hex (hashcat_ctx, line_buf, line_len);
+            get_next_word (hashcat_ctx_tmp, &fp, &line_buf, &line_len);
 
             // post-process rule engine
 
@@ -1486,6 +1406,14 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
               line_len = (u32) rule_len_out;
             }
 
+            if (user_options->attack_mode == ATTACK_MODE_ASSOCIATION)
+            {
+              // we can't reject password base on length in -a 9 because it will bring the schedule out of sync
+              // therefore we render it defective so the other candidates survive
+
+              line_len = MIN (line_len, hashconfig->pw_max);
+            }
+
             if (attack_kern == ATTACK_KERN_STRAIGHT)
             {
               if ((line_len < hashconfig->pw_min) || (line_len > hashconfig->pw_max))
@@ -1547,33 +1475,25 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
 
         if (pws_cnt)
         {
-          int CL_rc;
-
-          CL_rc = run_copy (hashcat_ctx, device_param, pws_cnt);
-
-          if (CL_rc == -1)
+          if (run_copy (hashcat_ctx, device_param, pws_cnt) == -1)
           {
-            if (attack_mode == ATTACK_MODE_COMBI) fclose (device_param->combs_fp);
+            if (attack_mode == ATTACK_MODE_COMBI) hc_fclose (&device_param->combs_fp);
 
-            fclose (fd);
+            hc_fclose (&fp);
 
             hcfree (hashcat_ctx_tmp->wl_data);
-
             hcfree (hashcat_ctx_tmp);
 
             return -1;
           }
 
-          CL_rc = run_cracker (hashcat_ctx, device_param, pws_cnt);
-
-          if (CL_rc == -1)
+          if (run_cracker (hashcat_ctx, device_param, device_param->words_off, pws_cnt) == -1)
           {
-            if (attack_mode == ATTACK_MODE_COMBI) fclose (device_param->combs_fp);
+            if (attack_mode == ATTACK_MODE_COMBI) hc_fclose (&device_param->combs_fp);
 
-            fclose (fd);
+            hc_fclose (&fp);
 
             hcfree (hashcat_ctx_tmp->wl_data);
-
             hcfree (hashcat_ctx_tmp);
 
             return -1;
@@ -1624,14 +1544,13 @@ static int calc (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param)
         if (words_fin == 0) break;
       }
 
-      if (attack_mode == ATTACK_MODE_COMBI) fclose (device_param->combs_fp);
+      if (attack_mode == ATTACK_MODE_COMBI) hc_fclose (&device_param->combs_fp);
 
-      fclose (fd);
+      hc_fclose (&fp);
 
       wl_data_destroy (hashcat_ctx_tmp);
 
       hcfree (hashcat_ctx_tmp->wl_data);
-
       hcfree (hashcat_ctx_tmp);
     }
   }
@@ -1650,18 +1569,22 @@ HC_API_CALL void *thread_calc (void *p)
   thread_param_t *thread_param = (thread_param_t *) p;
 
   hashcat_ctx_t *hashcat_ctx = thread_param->hashcat_ctx;
+  backend_ctx_t *backend_ctx = hashcat_ctx->backend_ctx;
 
-  opencl_ctx_t *opencl_ctx = hashcat_ctx->opencl_ctx;
+  if (backend_ctx->enabled == false) return NULL;
 
-  if (opencl_ctx->enabled == false) return NULL;
-
-  hc_device_param_t *device_param = opencl_ctx->devices_param + thread_param->tid;
+  hc_device_param_t *device_param = backend_ctx->devices_param + thread_param->tid;
 
   if (device_param->skipped) return NULL;
 
-  const int rc_calc = calc (hashcat_ctx, device_param);
+  if (device_param->skipped_warning == true) return NULL;
 
-  if (rc_calc == -1)
+  if (device_param->is_cuda == true)
+  {
+    if (hc_cuCtxSetCurrent (hashcat_ctx, device_param->cuda_context) == -1) return NULL;
+  }
+
+  if (calc (hashcat_ctx, device_param) == -1)
   {
     status_ctx_t *status_ctx = hashcat_ctx->status_ctx;
 
