@@ -8,25 +8,119 @@
 #include "memory.h"
 #include "event.h"
 #include "convert.h"
-#include "interface.h"
-#include "hashes.h"
 #include "mpsp.h"
 #include "rp.h"
-#include "rp_kernel_on_cpu.h"
-#include "rp_kernel_on_cpu_optimized.h"
-#include "opencl.h"
+#include "emu_inc_rp.h"
+#include "emu_inc_rp_optimized.h"
+#include "backend.h"
 #include "shared.h"
-#include "outfile.h"
 #include "locking.h"
+#include "outfile.h"
+
+u32 outfile_format_parse (const char *format_string)
+{
+  if (format_string == NULL) return 0;
+
+  char *format = hcstrdup (format_string);
+
+  if (format == NULL) return 0;
+
+  char *saveptr = NULL;
+
+  char *next = strtok_r (format, ",", &saveptr);
+
+  if (next == NULL)
+  {
+    hcfree (format);
+
+    return 0;
+  }
+
+  u32 outfile_format = 0;
+
+  do
+  {
+    const int tok_len = strlen (next);
+
+    // reject non-numbers:
+
+    if (is_valid_digit_string ((const u8 *) next, tok_len) == false)
+    {
+      outfile_format = 0;
+      break;
+    }
+
+    // string to number conversion:
+
+    const u32 num = hc_strtoul (next, NULL, 10);
+
+    if (num == 0)
+    {
+      outfile_format = 0;
+      break;
+    }
+
+    if (num > 31)
+    {
+      outfile_format = 0;
+      break;
+    }
+
+    // to bitmask:
+
+    const u32 bit = 1 << (num - 1);
+
+    bool accepted = false;
+
+    switch (bit)
+    {
+      // allowed formats:
+      case OUTFILE_FMT_HASH:
+      case OUTFILE_FMT_PLAIN:
+      case OUTFILE_FMT_HEXPLAIN:
+      case OUTFILE_FMT_CRACKPOS:
+      case OUTFILE_FMT_TIME_ABS:
+      case OUTFILE_FMT_TIME_REL:
+        accepted = true;
+        break;
+      // NOT acceptable formats:
+      default:
+        accepted = false;
+        break;
+    }
+
+    if (accepted == false)
+    {
+      outfile_format = 0;
+      break;
+    }
+
+    // the user should specify any format at most once:
+
+    if (outfile_format & bit)
+    {
+      outfile_format = 0;
+      break;
+    }
+
+    outfile_format |= bit;
+
+  } while ((next = strtok_r ((char *) NULL, ",", &saveptr)) != NULL);
+
+  hcfree (format);
+
+  return outfile_format;
+}
 
 int build_plain (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param, plain_t *plain, u32 *plain_buf, int *out_len)
 {
-  const combinator_ctx_t *combinator_ctx = hashcat_ctx->combinator_ctx;
-  const hashconfig_t     *hashconfig     = hashcat_ctx->hashconfig;
-  const hashes_t         *hashes         = hashcat_ctx->hashes;
-  const mask_ctx_t       *mask_ctx       = hashcat_ctx->mask_ctx;
-  const straight_ctx_t   *straight_ctx   = hashcat_ctx->straight_ctx;
-  const user_options_t   *user_options   = hashcat_ctx->user_options;
+  const combinator_ctx_t     *combinator_ctx     = hashcat_ctx->combinator_ctx;
+  const hashconfig_t         *hashconfig         = hashcat_ctx->hashconfig;
+  const hashes_t             *hashes             = hashcat_ctx->hashes;
+  const mask_ctx_t           *mask_ctx           = hashcat_ctx->mask_ctx;
+  const straight_ctx_t       *straight_ctx       = hashcat_ctx->straight_ctx;
+  const user_options_t       *user_options       = hashcat_ctx->user_options;
+  const user_options_extra_t *user_options_extra = hashcat_ctx->user_options_extra;
 
   const u64 gidvid = plain->gidvid;
   const u32 il_pos = plain->il_pos;
@@ -49,7 +143,7 @@ int build_plain (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param, pl
   }
   else
   {
-    if (user_options->attack_mode == ATTACK_MODE_STRAIGHT)
+    if ((user_options->attack_mode == ATTACK_MODE_STRAIGHT) || (user_options->attack_mode == ATTACK_MODE_ASSOCIATION))
     {
       pw_t pw;
 
@@ -61,12 +155,24 @@ int build_plain (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param, pl
 
       if (hashconfig->opti_type & OPTI_TYPE_OPTIMIZED_KERNEL)
       {
-        for (int i = 0; i < 8; i++)
+        if ((user_options->rp_files_cnt == 0) && (user_options->rp_gen == 0))
         {
-          plain_buf[i] = pw.i[i];
-        }
+          for (int i = 0; i < 14; i++)
+          {
+            plain_buf[i] = pw.i[i];
+          }
 
-        plain_len = apply_rules_optimized (straight_ctx->kernel_rules_buf[off].cmds, &plain_buf[0], &plain_buf[4], pw.pw_len);
+          plain_len = pw.pw_len;
+        }
+        else
+        {
+          for (int i = 0; i < 8; i++)
+          {
+            plain_buf[i] = pw.i[i];
+          }
+
+          plain_len = apply_rules_optimized (straight_ctx->kernel_rules_buf[off].cmds, &plain_buf[0], &plain_buf[4], pw.pw_len);
+        }
       }
       else
       {
@@ -237,9 +343,30 @@ int build_plain (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param, pl
     }
   }
 
-  const u32 pw_max = hashconfig_get_pw_max (hashcat_ctx, false);
+  int pw_max = (const int) hashconfig->pw_max;
 
-  if (plain_len > (int) hashconfig->pw_max) plain_len = MIN (plain_len, (int) pw_max);
+  // pw_max is per pw_t element but in combinator we have two pw_t elements.
+  // therefore we can support up to 64 in combinator in optimized mode (but limited by general hash limit 55)
+  // or full 512 in pure mode (but limited by hashcat buffer size limit 256).
+  // some algorithms do not support general default pw_max = 31,
+  // therefore we need to use pw_max as a base and not hardcode it.
+
+  if (plain_len > pw_max)
+  {
+    if (user_options_extra->attack_kern == ATTACK_KERN_COMBI)
+    {
+      if (hashconfig->opti_type & OPTI_TYPE_OPTIMIZED_KERNEL)
+      {
+        pw_max = MIN ((pw_max * 2), 55);
+      }
+      else
+      {
+        pw_max = MIN ((pw_max * 2), 256);
+      }
+    }
+  }
+
+  if (plain_len > pw_max) plain_len = MIN (plain_len, pw_max);
 
   plain_ptr[plain_len] = 0;
 
@@ -301,7 +428,7 @@ int build_debugdata (hashcat_ctx_t *hashcat_ctx, hc_device_param_t *device_param
   const u64 gidvid = plain->gidvid;
   const u32 il_pos = plain->il_pos;
 
-  if (user_options->attack_mode != ATTACK_MODE_STRAIGHT) return 0;
+  if ((user_options->attack_mode != ATTACK_MODE_STRAIGHT) && (user_options->attack_mode != ATTACK_MODE_ASSOCIATION)) return 0;
 
   const u32 debug_mode = debugfile_ctx->mode;
 
@@ -372,10 +499,10 @@ int outfile_init (hashcat_ctx_t *hashcat_ctx)
   outfile_ctx_t  *outfile_ctx  = hashcat_ctx->outfile_ctx;
   user_options_t *user_options = hashcat_ctx->user_options;
 
-  outfile_ctx->fp               = NULL;
-  outfile_ctx->filename         = user_options->outfile;
-  outfile_ctx->outfile_format   = user_options->outfile_format;
-  outfile_ctx->outfile_autohex  = user_options->outfile_autohex;
+  outfile_ctx->fp.pfp          = NULL;
+  outfile_ctx->filename        = user_options->outfile;
+  outfile_ctx->outfile_format  = user_options->outfile_format;
+  outfile_ctx->outfile_autohex = user_options->outfile_autohex;
 
   return 0;
 }
@@ -393,26 +520,22 @@ int outfile_write_open (hashcat_ctx_t *hashcat_ctx)
 
   if (outfile_ctx->filename == NULL) return 0;
 
-  FILE *fp = fopen (outfile_ctx->filename, "ab");
-
-  if (fp == NULL)
+  if (hc_fopen (&outfile_ctx->fp, outfile_ctx->filename, "ab") == false)
   {
     event_log_error (hashcat_ctx, "%s: %s", outfile_ctx->filename, strerror (errno));
 
     return -1;
   }
 
-  if (lock_file (fp) == -1)
+  if (hc_lockfile (&outfile_ctx->fp) == -1)
   {
-    fclose (fp);
+    hc_fclose (&outfile_ctx->fp);
 
     event_log_error (hashcat_ctx, "%s: %s", outfile_ctx->filename, strerror (errno));
 
     return -1;
   }
 
-  outfile_ctx->fp = fp;
-
   return 0;
 }
 
@@ -420,16 +543,21 @@ void outfile_write_close (hashcat_ctx_t *hashcat_ctx)
 {
   outfile_ctx_t *outfile_ctx = hashcat_ctx->outfile_ctx;
 
-  if (outfile_ctx->fp == NULL) return;
+  if (outfile_ctx->fp.pfp == NULL) return;
 
-  fclose (outfile_ctx->fp);
+  hc_unlockfile (&outfile_ctx->fp);
+
+  hc_fclose (&outfile_ctx->fp);
 }
 
-int outfile_write (hashcat_ctx_t *hashcat_ctx, const char *out_buf, const unsigned char *plain_ptr, const u32 plain_len, const u64 crackpos, const unsigned char *username, const u32 user_len, char tmp_buf[HCBUFSIZ_LARGE])
+int outfile_write (hashcat_ctx_t *hashcat_ctx, const char *out_buf, const int out_len, const unsigned char *plain_ptr, const u32 plain_len, const u64 crackpos, const unsigned char *username, const u32 user_len, const bool print_eol, char tmp_buf[HCBUFSIZ_LARGE])
 {
   const hashconfig_t   *hashconfig   = hashcat_ctx->hashconfig;
-  const outfile_ctx_t  *outfile_ctx  = hashcat_ctx->outfile_ctx;
   const user_options_t *user_options = hashcat_ctx->user_options;
+  outfile_ctx_t        *outfile_ctx  = hashcat_ctx->outfile_ctx;
+  status_ctx_t         *status_ctx   = hashcat_ctx->status_ctx;
+
+  const u32 outfile_format = (hashconfig->opts_type & OPTS_TYPE_PT_ALWAYS_HEXIFY) ? 5 : outfile_ctx->outfile_format;
 
   int tmp_len = 0;
 
@@ -441,7 +569,7 @@ int outfile_write (hashcat_ctx_t *hashcat_ctx, const char *out_buf, const unsign
 
       tmp_len += user_len;
 
-      if (outfile_ctx->outfile_format & (OUTFILE_FMT_HASH | OUTFILE_FMT_PLAIN | OUTFILE_FMT_HEXPLAIN | OUTFILE_FMT_CRACKPOS))
+      if (outfile_format & (OUTFILE_FMT_TIME_ABS | OUTFILE_FMT_TIME_REL | OUTFILE_FMT_HASH | OUTFILE_FMT_PLAIN | OUTFILE_FMT_HEXPLAIN | OUTFILE_FMT_CRACKPOS))
       {
         tmp_buf[tmp_len] = hashconfig->separator;
 
@@ -450,15 +578,15 @@ int outfile_write (hashcat_ctx_t *hashcat_ctx, const char *out_buf, const unsign
     }
   }
 
-  if (outfile_ctx->outfile_format & OUTFILE_FMT_HASH)
+  if (outfile_format & OUTFILE_FMT_TIME_ABS)
   {
-    const size_t out_len = strlen (out_buf);
+    time_t now;
 
-    memcpy (tmp_buf + tmp_len, out_buf, out_len);
+    time (&now);
 
-    tmp_len += out_len;
+    tmp_len += snprintf (tmp_buf + tmp_len, HCBUFSIZ_LARGE - tmp_len, "%" PRIu64, (u64) now);
 
-    if (outfile_ctx->outfile_format & (OUTFILE_FMT_PLAIN | OUTFILE_FMT_HEXPLAIN | OUTFILE_FMT_CRACKPOS))
+    if (outfile_format & (OUTFILE_FMT_TIME_REL | OUTFILE_FMT_HASH | OUTFILE_FMT_PLAIN | OUTFILE_FMT_HEXPLAIN | OUTFILE_FMT_CRACKPOS))
     {
       tmp_buf[tmp_len] = hashconfig->separator;
 
@@ -466,7 +594,46 @@ int outfile_write (hashcat_ctx_t *hashcat_ctx, const char *out_buf, const unsign
     }
   }
 
-  if (outfile_ctx->outfile_format & OUTFILE_FMT_PLAIN)
+  if (outfile_format & OUTFILE_FMT_TIME_REL)
+  {
+    time_t time_now;
+
+    time (&time_now);
+
+    time_t time_started = status_ctx->runtime_start;
+
+    u64 diff = 0;
+
+    if (time_now > time_started) // should always be true, but you never know
+    {
+      diff = (u64) time_now - (u64) time_started;
+    }
+
+    tmp_len += snprintf (tmp_buf + tmp_len, HCBUFSIZ_LARGE - tmp_len, "%" PRIu64, diff);
+
+    if (outfile_format & (OUTFILE_FMT_HASH | OUTFILE_FMT_PLAIN | OUTFILE_FMT_HEXPLAIN | OUTFILE_FMT_CRACKPOS))
+    {
+      tmp_buf[tmp_len] = hashconfig->separator;
+
+      tmp_len += 1;
+    }
+  }
+
+  if (outfile_format & OUTFILE_FMT_HASH)
+  {
+    memcpy (tmp_buf + tmp_len, out_buf, out_len);
+
+    tmp_len += out_len;
+
+    if (outfile_format & (OUTFILE_FMT_PLAIN | OUTFILE_FMT_HEXPLAIN | OUTFILE_FMT_CRACKPOS))
+    {
+      tmp_buf[tmp_len] = hashconfig->separator;
+
+      tmp_len += 1;
+    }
+  }
+
+  if (outfile_format & OUTFILE_FMT_PLAIN)
   {
     bool convert_to_hex = false;
 
@@ -474,7 +641,7 @@ int outfile_write (hashcat_ctx_t *hashcat_ctx, const char *out_buf, const unsign
     {
       if (user_options->outfile_autohex == true)
       {
-        const bool always_ascii = (hashconfig->hash_type & OPTS_TYPE_PT_ALWAYS_ASCII) ? true : false;
+        const bool always_ascii = (hashconfig->opts_type & OPTS_TYPE_PT_ALWAYS_ASCII) ? true : false;
 
         convert_to_hex = need_hexify (plain_ptr, plain_len, hashconfig->separator, always_ascii);
       }
@@ -501,7 +668,7 @@ int outfile_write (hashcat_ctx_t *hashcat_ctx, const char *out_buf, const unsign
       tmp_len += plain_len;
     }
 
-    if (outfile_ctx->outfile_format & (OUTFILE_FMT_HEXPLAIN | OUTFILE_FMT_CRACKPOS))
+    if (outfile_format & (OUTFILE_FMT_HEXPLAIN | OUTFILE_FMT_CRACKPOS))
     {
       tmp_buf[tmp_len] = hashconfig->separator;
 
@@ -509,13 +676,13 @@ int outfile_write (hashcat_ctx_t *hashcat_ctx, const char *out_buf, const unsign
     }
   }
 
-  if (outfile_ctx->outfile_format & OUTFILE_FMT_HEXPLAIN)
+  if (outfile_format & OUTFILE_FMT_HEXPLAIN)
   {
     exec_hexify (plain_ptr, plain_len, (u8 *) tmp_buf + tmp_len);
 
     tmp_len += plain_len * 2;
 
-    if (outfile_ctx->outfile_format & (OUTFILE_FMT_CRACKPOS))
+    if (outfile_format & (OUTFILE_FMT_CRACKPOS))
     {
       tmp_buf[tmp_len] = hashconfig->separator;
 
@@ -523,17 +690,21 @@ int outfile_write (hashcat_ctx_t *hashcat_ctx, const char *out_buf, const unsign
     }
   }
 
-  if (outfile_ctx->outfile_format & OUTFILE_FMT_CRACKPOS)
+  if (outfile_format & OUTFILE_FMT_CRACKPOS)
   {
     tmp_len += snprintf (tmp_buf + tmp_len, HCBUFSIZ_LARGE - tmp_len, "%" PRIu64, crackpos);
   }
 
   tmp_buf[tmp_len] = 0;
 
-  if (outfile_ctx->fp != NULL)
+  if (outfile_ctx->fp.pfp != NULL)
   {
-    hc_fwrite (tmp_buf, tmp_len,      1, outfile_ctx->fp);
-    hc_fwrite (EOL,     strlen (EOL), 1, outfile_ctx->fp);
+    hc_fwrite (tmp_buf, tmp_len, 1, &outfile_ctx->fp);
+
+    if (print_eol == true)
+    {
+      hc_fwrite (EOL, strlen (EOL), 1, &outfile_ctx->fp);
+    }
   }
 
   return tmp_len;
