@@ -5,17 +5,11 @@
 
 #include "common.h"
 #include "types.h"
-#include "interface.h"
-#include "timer.h"
 #include "event.h"
 #include "memory.h"
 #include "filehandling.h"
-#include "ext_OpenCL.h"
-#include "tuningdb.h"
-#include "thread.h"
-#include "opencl.h"
-#include "hashes.h"
 #include "shared.h"
+#include "tuningdb.h"
 
 static int sort_by_tuning_db_alias (const void *v1, const void *v2)
 {
@@ -43,8 +37,8 @@ static int sort_by_tuning_db_entry (const void *v1, const void *v2)
 
   if (res2 != 0) return (res2);
 
-  const int res3 = t1->hash_type
-                 - t2->hash_type;
+  const int res3 = t1->hash_mode
+                 - t2->hash_mode;
 
   if (res3 != 0) return (res3);
 
@@ -60,10 +54,10 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
 
   tuning_db->enabled = false;
 
-  if (user_options->example_hashes == true) return 0;
+  if (user_options->hash_info      == true) return 0;
   if (user_options->keyspace       == true) return 0;
   if (user_options->left           == true) return 0;
-  if (user_options->opencl_info    == true) return 0;
+  if (user_options->backend_info   == true) return 0;
   if (user_options->show           == true) return 0;
   if (user_options->usage          == true) return 0;
   if (user_options->version        == true) return 0;
@@ -74,9 +68,9 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
 
   hc_asprintf (&tuning_db_file, "%s/%s", folder_config->shared_dir, TUNING_DB_FILE);
 
-  FILE *fp = fopen (tuning_db_file, "rb");
+  HCFILE fp;
 
-  if (fp == NULL)
+  if (hc_fopen (&fp, tuning_db_file, "rb") == false)
   {
     event_log_error (hashcat_ctx, "%s: %s", tuning_db_file, strerror (errno));
 
@@ -85,7 +79,7 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
 
   hcfree (tuning_db_file);
 
-  int num_lines = count_lines (fp);
+  const size_t num_lines = count_lines (&fp);
 
   // a bit over-allocated
 
@@ -95,21 +89,21 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
   tuning_db->entry_buf = (tuning_db_entry_t *) hccalloc (num_lines + 1, sizeof (tuning_db_entry_t));
   tuning_db->entry_cnt = 0;
 
-  rewind (fp);
+  hc_rewind (&fp);
 
   int line_num = 0;
 
   char *buf = (char *) hcmalloc (HCBUFSIZ_LARGE);
 
-  while (!feof (fp))
+  while (!hc_feof (&fp))
   {
-    char *line_buf = fgets (buf, HCBUFSIZ_LARGE - 1, fp);
+    char *line_buf = hc_fgets (buf, HCBUFSIZ_LARGE - 1, &fp);
 
     if (line_buf == NULL) break;
 
     line_num++;
 
-    const int line_len = in_superchop (line_buf);
+    const size_t line_len = in_superchop (line_buf);
 
     if (line_len == 0) continue;
 
@@ -121,7 +115,7 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
 
     int token_cnt = 0;
 
-    char *saveptr;
+    char *saveptr = NULL;
 
     char *next = strtok_r (line_buf, "\t ", &saveptr);
 
@@ -129,7 +123,7 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
 
     token_cnt++;
 
-    while ((next = strtok_r (NULL, "\t ", &saveptr)) != NULL)
+    while ((next = strtok_r ((char *) NULL, "\t ", &saveptr)) != NULL)
     {
       token_ptr[token_cnt] = next;
 
@@ -153,6 +147,7 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
       if ((token_ptr[1][0] != '0') &&
           (token_ptr[1][0] != '1') &&
           (token_ptr[1][0] != '3') &&
+          (token_ptr[1][0] != '9') &&
           (token_ptr[1][0] != '*'))
       {
         event_log_warning (hashcat_ctx, "Tuning-db: Invalid attack_mode '%c' in Line '%d'", token_ptr[1][0], line_num);
@@ -174,14 +169,14 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
       char *device_name = token_ptr[0];
 
       int attack_mode      = -1;
-      int hash_type        = -1;
+      int hash_mode        = -1;
       int vector_width     = -1;
       int kernel_accel     = -1;
       int kernel_loops     = -1;
 
-      if (token_ptr[1][0] != '*') attack_mode      = atoi (token_ptr[1]);
-      if (token_ptr[2][0] != '*') hash_type        = atoi (token_ptr[2]);
-      if (token_ptr[3][0] != 'N') vector_width     = atoi (token_ptr[3]);
+      if (token_ptr[1][0] != '*') attack_mode   = (int) strtol (token_ptr[1], NULL, 10);
+      if (token_ptr[2][0] != '*') hash_mode     = (int) strtol (token_ptr[2], NULL, 10);
+      if (token_ptr[3][0] != 'N') vector_width  = (int) strtol (token_ptr[3], NULL, 10);
 
       if (token_ptr[4][0] == 'A')
       {
@@ -191,9 +186,13 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
       {
         kernel_accel = 1024;
       }
+      else if (token_ptr[4][0] == 'N')
+      {
+        kernel_accel = -1;
+      }
       else
       {
-        kernel_accel = atoi (token_ptr[4]);
+        kernel_accel = (int) strtol (token_ptr[4], NULL, 10);
 
         if ((kernel_accel < 1) || (kernel_accel > 1024))
         {
@@ -224,7 +223,7 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
       }
       else
       {
-        kernel_loops = atoi (token_ptr[5]);
+        kernel_loops = (int) strtol (token_ptr[5], NULL, 10);
 
         if (kernel_loops < 1)
         {
@@ -259,7 +258,7 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
 
       entry->device_name  = hcstrdup (device_name);
       entry->attack_mode  = attack_mode;
-      entry->hash_type    = hash_type;
+      entry->hash_mode    = hash_mode;
       entry->vector_width = vector_width;
       entry->kernel_accel = kernel_accel;
       entry->kernel_loops = kernel_loops;
@@ -276,7 +275,7 @@ int tuning_db_init (hashcat_ctx_t *hashcat_ctx)
 
   hcfree (buf);
 
-  fclose (fp);
+  hc_fclose (&fp);
 
   // todo: print loaded 'cnt' message
 
@@ -317,7 +316,7 @@ void tuning_db_destroy (hashcat_ctx_t *hashcat_ctx)
   memset (tuning_db, 0, sizeof (tuning_db_t));
 }
 
-tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *device_name, const cl_device_type device_type, int attack_mode, const int hash_type)
+tuning_db_entry_t *tuning_db_search_real (hashcat_ctx_t *hashcat_ctx, const char *device_name, const cl_device_type device_type, int attack_mode, const int hash_mode)
 {
   tuning_db_t *tuning_db = hashcat_ctx->tuning_db;
 
@@ -327,9 +326,9 @@ tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *dev
 
   char *device_name_nospace = hcstrdup (device_name);
 
-  int device_name_length = strlen (device_name_nospace);
+  const size_t device_name_length = strlen (device_name_nospace);
 
-  int i;
+  size_t i;
 
   for (i = 0; i < device_name_length; i++)
   {
@@ -338,17 +337,19 @@ tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *dev
 
   // find out if there's an alias configured
 
+  char *device_name_nospace2 = hcstrdup (device_name_nospace);
+
   tuning_db_alias_t a;
 
-  a.device_name = device_name_nospace;
+  a.device_name = device_name_nospace2;
 
   char *alias_name = NULL;
 
   for (i = device_name_length; i >= 1; i--)
   {
-    device_name_nospace[i] = 0;
+    device_name_nospace2[i] = 0;
 
-    tuning_db_alias_t *alias = bsearch (&a, tuning_db->alias_buf, tuning_db->alias_cnt, sizeof (tuning_db_alias_t), sort_by_tuning_db_alias);
+    tuning_db_alias_t *alias = (tuning_db_alias_t *) bsearch (&a, tuning_db->alias_buf, tuning_db->alias_cnt, sizeof (tuning_db_alias_t), sort_by_tuning_db_alias);
 
     if (alias == NULL) continue;
 
@@ -357,6 +358,8 @@ tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *dev
     break;
   }
 
+  hcfree (device_name_nospace2);
+
   // attack-mode 6 and 7 are attack-mode 1 basically
 
   if (attack_mode == 6) attack_mode = 1;
@@ -366,7 +369,7 @@ tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *dev
 
   s.device_name = device_name_nospace;
   s.attack_mode = attack_mode;
-  s.hash_type   = hash_type;
+  s.hash_mode   = hash_mode;
 
   tuning_db_entry_t *entry = NULL;
 
@@ -376,9 +379,9 @@ tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *dev
   {
     s.device_name = (i & 1) ? "*" : device_name_nospace;
     s.attack_mode = (i & 2) ?  -1 : attack_mode;
-    s.hash_type   = (i & 4) ?  -1 : hash_type;
+    s.hash_mode   = (i & 4) ?  -1 : hash_mode;
 
-    entry = bsearch (&s, tuning_db->entry_buf, tuning_db->entry_cnt, sizeof (tuning_db_entry_t), sort_by_tuning_db_entry);
+    entry = (tuning_db_entry_t *) bsearch (&s, tuning_db->entry_buf, tuning_db->entry_cnt, sizeof (tuning_db_entry_t), sort_by_tuning_db_entry);
 
     if (entry != NULL) break;
 
@@ -392,7 +395,7 @@ tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *dev
       {
         s.device_name = alias_name;
 
-        entry = bsearch (&s, tuning_db->entry_buf, tuning_db->entry_cnt, sizeof (tuning_db_entry_t), sort_by_tuning_db_entry);
+        entry = (tuning_db_entry_t *) bsearch (&s, tuning_db->entry_buf, tuning_db->entry_cnt, sizeof (tuning_db_entry_t), sort_by_tuning_db_entry);
 
         if (entry != NULL) break;
       }
@@ -412,7 +415,7 @@ tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *dev
         s.device_name = "DEVICE_TYPE_ACCELERATOR";
       }
 
-      entry = bsearch (&s, tuning_db->entry_buf, tuning_db->entry_cnt, sizeof (tuning_db_entry_t), sort_by_tuning_db_entry);
+      entry = (tuning_db_entry_t *) bsearch (&s, tuning_db->entry_buf, tuning_db->entry_cnt, sizeof (tuning_db_entry_t), sort_by_tuning_db_entry);
 
       if (entry != NULL) break;
     }
@@ -424,3 +427,21 @@ tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *dev
 
   return entry;
 }
+
+tuning_db_entry_t *tuning_db_search (hashcat_ctx_t *hashcat_ctx, const char *device_name, const cl_device_type device_type, int attack_mode, const int hash_mode)
+{
+  tuning_db_entry_t *entry;
+
+  const char *NV_prefix = (const char *) "NVIDIA ";
+
+  if (memcmp (device_name, NV_prefix, strlen (NV_prefix)) == 0)
+  {
+    entry = tuning_db_search_real (hashcat_ctx, device_name + strlen (NV_prefix), device_type, attack_mode, hash_mode);
+
+    if (entry) return entry;
+  }
+
+  entry = tuning_db_search_real (hashcat_ctx, device_name, device_type, attack_mode, hash_mode);
+
+  return entry;
+}
